#!/usr/bin/env python2
# -*- coding: utf-8 -*-
from __future__ import print_function
import sys
import re


"""parse-valgrind.py: valgrind log visualizer"""
__version__   = "1.1"
__author__    = "ed <parsevalgrind@ocv.me>"
__credits__   = ["stackoverflow.com"]
__license__   = "MIT"
__copyright__ = 2017


# start:
# valgrind --tool=memcheck --trace-children=yes --log-file=/tmp/valgrind.log --xml=yes --xml-file=/tmp/valgrind.xml --leak-check=full --show-reachable=yes ANALYSIS_TARGET_HERE
# valgrind --tool=helgrind --trace-children=yes --log-file=/tmp/helgrind.log --xml=yes --xml-file=/tmp/helgrind.xml ANALYSIS_TARGET_HERE
#
# stop and parse:
# kill -TERM $(ps aux | grep 'bin[/]valgrind ' | awk '{print $2}'); while true; do ps aux | grep -qE 'bin[/]valgrind' || break; sleep 1; done; parse-valgrind.py /tmp/valgrind.xml
#
#
#
# running valgrind with support for gdb (vgdb-error=0 breaks on all illegals):
# valgrind --vgdb=yes --vgdb-error=0 ANALYSIS_TARGET_HERE
#
# attaching gdb to inspect and resume from breakpoint:
# gdb -ex 'target remote | vgdb' -ex bt -ex c ANALYSIS_TARGET_HERE
#
# collect a stacktrace and break:
# gdb -ex 'target remote | vgdb' -ex 'set height 0' -ex 't a a bt' ANALYSIS_TARGET_HERE
#
# terminating valgrind with gdb:
# gdb -ex 'target remote | vgdb' -ex 'monitor v.kill' ANALYSIS_TARGET_HERE


def termsize():
	import os
	env = os.environ
	def ioctl_GWINSZ(fd):
		try:
			import fcntl, termios, struct, os
			cr = struct.unpack('hh', fcntl.ioctl(fd, termios.TIOCGWINSZ, '1234'))
		except:
			return
		return cr
	cr = ioctl_GWINSZ(0) or ioctl_GWINSZ(1) or ioctl_GWINSZ(2)
	if not cr:
		try:
			fd = os.open(os.ctermid(), os.O_RDONLY)
			cr = ioctl_GWINSZ(fd)
			os.close(fd)
		except:
			pass
	if not cr:
		try:
			cr = (env['LINES'], env['COLUMNS'])
		except:
			cr = (25, 80)
	return int(cr[1]), int(cr[0])
sz = termsize()
sw = sz[0]
sh = sz[1]



def xmldec(s):
	return (s.
		replace('<br/>',  '\n').
		replace('&#9;',   '    ').
		replace('&quot;', '"').
		replace('&gt;',   '>').
		replace('&lt;',   '<').
		replace('&amp;',  '&'))



rx = re.compile('^( *)<([^>]*)>(.*)</[^>]*>$')

class Event(object):
	def __init__(self, uniq, kind, size, snapshots):
		self.uniq = uniq
		self.kind = kind
		self.size = size
		self.snapshots = snapshots
		self.hits = 1

class Snapshot(object):
	def __init__(self, desc, stack):
		self.stack = stack
		self.desc = desc

def trunc(path):
	nodes = path.split('/')
	if len(nodes) < 6:
		return path
	return '{0}/{1}/{2}/.../{3}/{4}'.format(*(nodes[:3] + nodes[-2:]))

def main():
	if len(sys.argv) != 2:
		print('need argument 1: valgrind xml to parse')
		sys.exit(1)
	
	pause_on = [      '<args>',  '<preamble>',  '<status>',  '<fatal_signal>',  '<suppcounts>',  '<announcethread>' ]
	resume_after = [ '</args>', '</preamble>', '</status>', '</fatal_signal>', '</suppcounts>', '</announcethread>' ]
	ignored_keys = [ 'tid', 'dir', 'leakedblocks', 'hthreadid' ]

	events = []
	active = False
	in_errcount = False
	
	num_lines = 0
	with open(sys.argv[1], 'rb') as f:
		for line in f:
			num_lines += 1

	print()
	dbg_on = False
	lines_read = 0
	print_step = max(int(num_lines / 100), 1)
	with open(sys.argv[1], 'rb') as f:
		for line in [x.rstrip() for x in f]:
			lines_read += 1
			if lines_read % print_step == 0:
				print('\033[Aparsing report, {0}% done, line {1}, events {2} '.format(
					int(lines_read/print_step), lines_read, len(events)))
			
			#dbg_on = lines_read > 6535 and lines_read < 6570
			#if dbg_on:
			#	print('\033[1;30m' + line + '\033[0m')
			
			if line == '<errorcounts>':
				in_errcount = True
			
			if line == '</errorcounts>':
				in_errcount = False
			
			if in_errcount:
				if '<pair>' in line:
					uniq = None
					count = 0
					continue
				
				if '</pair>' in line:
					for event in events:
						if event.uniq == uniq:
							#if event.kind.startswith('Leak_'):
							#	raise RuntimeError('not implmeneted: multiplication of event.desc: {0}, {1}, {2}, {3}'.format(
							#		event, event.kind, event.snapshots[0].desc,
							#		'{0} * {1} = {2}'.format(event.size, count, event.size * count)))
							event.hits = count
							event.size *= count
							event.snapshots[0].desc = '\033[1;37m{0}x \033[1;35m{1} \033[30m{2}'.format(
								event.hits, event.snapshots[0].desc, event.kind)
							break
					continue
				
				m = rx.search(line)
				if not m:
					continue
				
				xindent, xkey, xval = m.groups()
				xval = xmldec(xval)
				if xkey == 'unique':
					uniq = xval
				elif xkey == 'count':
					count = int(xval)
				
				continue
			
			if not active:
				if line == '<error>':
					active = True
				elif events and line in resume_after:
					active = True
					continue
				else:
					continue
			
			if line in pause_on:
				active = False
				continue
			
			if line == '    <frame>':
				ip   = '?'
				obj  = '?'
				fn   = '?'
				filn = '?'
				flno = '?'
				continue
			
			if line == '<error>':
				desc = ''
				stack = []
				snapshots = []
				size = None
				uniq = None
				continue
			
			if line == '    </frame>':
				pad = 11+42+5
				w = sw - pad

				obj_trunc = trunc(obj)
				overshoot = len(obj_trunc) - 42
				if overshoot > 0:
					w -= overshoot
				
				wrapped = u''
				wraps = u''
				while len(fn) > w:
					wrapped += fn[:w-1] + u'┓'
					fn = u'┗━' + fn[w-1:]
					wrapped += u'\n' + u' ' * pad
					wraps = u'\n'
				
				wrapped += fn
				if len(fn) > w - (len(filn + flno) + 6):
					wrapped += u'\n{0}┗━'.format(u' ' * pad)
					wraps = u'\n'
				
				stack.append(u'\033[1;31m{0:>12}\033[0m in \033[1;33m{1:42}\033[0m \033[1;32m{2}\033[0m @ \033[1;34m{3}:{4}\033[0m{5}'.format(
					ip, obj_trunc, wrapped, filn, flno, wraps))
				continue
			
			if line == '</error>' \
			or '  <auxwhat>' in line \
			or '  <xauxwhat>' in line:
				snapshots.append(Snapshot(desc, stack))
				desc = ''
				stack = []
			
			if line == '</error>':
				if size is None:
					# valgrind
					if kind == 'SyscallParam':
						size = 100000000000
					elif kind == 'InvalidJump':
						size = 10000000000
					elif kind == 'InvalidWrite':
						size = 10000000000
					elif kind == 'InvalidFree':
						size = 1000000000
					elif kind == 'MismatchedFree':
						size = 1000000000
					elif kind == 'InvalidRead':
						size = 100000000
					elif kind == 'UninitValue':
						size = 1000000
					elif kind == 'UninitCondition':
						size = 1000000
					
					# helgrind
					elif kind == 'LockOrder':
						size = 1000000000000
					elif kind == 'UnlockForeign':
						size = 10000000000
					elif kind == 'Race':
						size = 100000000
					elif kind == 'Misc':
						size = 1000000
					
					else:
						raise RuntimeError('\033[1;31munknown error kind:\033[0m  ' + kind)
				
				events.append(Event(uniq, kind, size, snapshots))
				continue
			
			m = rx.search(line)
			if not m:
				continue
			
			xindent, xkey, xval = m.groups()
			xval = xmldec(xval)
			
			if xkey == 'kind':
				kind = xval
			
			elif xkey == 'what':
				desc = xval
			
			elif xkey == 'what':
				desc = xval
			
			elif xkey == 'text':
				desc = xval
			
			elif xkey == 'auxwhat':
				desc = xval
			
			elif xkey == 'unique':
				uniq = xval
			
			elif xkey == 'leakedbytes':
				size = int(xval)
			
			elif xkey == 'ip':
				ip = xval
			
			elif xkey == 'obj':
				obj = xval
			
			elif xkey == 'fn':
				fn = xval
			
			elif xkey == 'file':
				filn = xval
			
			elif xkey == 'line':
				flno = xval
			
			elif xkey in ignored_keys:
				continue
			
			else:
				raise RuntimeError('\033[1;31munknown syntax in log entry:\033[0m  ' + line)


	#return
	for n in range(sh*5):
		print()
	

	for event in sorted(events, key=lambda x: x.size):

		# skip leaks smaller than 1KB cumulative
		if (event.size < 1024):
			continue
		
		# skip if L::~L() is mentioned in any stack, any frame
		if any(any('L::~L()' in x for x in y.stack) for y in event.snapshots):
			continue
		
		# additional example filters
		if False:
			# skip if libfoo.so mentioned in the 3rd stack frame
			if '/libfoo.so' in event.snapshots[0].stack[:3][-1]:
				continue
		
			# skip if the event is an unitialized condition
			# and libfoo.so is mentioned in the topmost stack frame
			if event.kind == 'UninitCondition' \
			and '/libfoo.so' in event.snapshots[0].stack[0]:
				continue
		
			# skip if libfoo.so appears in 80% of the stack's frames:
			n_hits = 0
			for v in event.snapshots[0].stack:
				if '/libfoo.so' in v:
					n_hits += 1
			
			if n_hits * 1. / len(event.snapshots[0].stack) >= 0.8:
				continue

		color = '\033[1;35m'

		for snapshot in event.snapshots:
			
			if snapshot.desc:
				header = snapshot.desc
				if not header.startswith('\033'):
					header = color + header
				color = '\033[1;33m┗━\033[1;37m'
				
				print(header)
			
			if snapshot.stack:
				print('\n'.join(snapshot.stack))
				epilogue = ''
			else:
				epilogue = '\n'
		
		print('{0}\033[0;33m{1}\033[0m\n'.format(
			epilogue, ('━'*sw)))



main()
print('done')
